// SPDX-License-Identifier: GPL-2.0
/*
 * FB driver for the NHD-1.69-160128UGC3 (Newhaven Display International, Inc.)
 * using the SEPS525 (Syncoam) LCD Controller
 *
 * Copyright (C) 2016 Analog Devices Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A spiTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#define DEBUG

#include <common.h>
#include <command.h>
#include <cpu_func.h>
#include <errno.h>
#include <linux/delay.h>
#include <asm/gpio.h>
#include <spi.h>
#include <dm.h>
#include <video.h>
#include <asm/gpio.h>

#define WIDTH		160
#define HEIGHT		128

#define CONFIG_SEPS525_BUS 0
#define CONFIG_SEPS525_CS  0
#define CONFIG_SEPS525_RS  27

#define SEPS525_INDEX 0x00
#define SEPS525_STATUS_RD 0x01
#define SEPS525_OSC_CTL 0x02
#define SEPS525_IREF 0x80
#define SEPS525_CLOCK_DIV 0x03
#define SEPS525_REDUCE_CURRENT 0x04
#define SEPS525_SOFT_RST 0x05
#define SEPS525_DISP_ONOFF 0x06
#define SEPS525_PRECHARGE_TIME_R 0x08
#define SEPS525_PRECHARGE_TIME_G 0x09
#define SEPS525_PRECHARGE_TIME_B 0x0A
#define SEPS525_PRECHARGE_CURRENT_R 0x0B
#define SEPS525_PRECHARGE_CURRENT_G 0x0C
#define SEPS525_PRECHARGE_CURRENT_B 0x0D
#define SEPS525_DRIVING_CURRENT_R 0x10
#define SEPS525_DRIVING_CURRENT_G 0x11
#define SEPS525_DRIVING_CURRENT_B 0x12
#define SEPS525_DISPLAYMODE_SET 0x13
#define SEPS525_RGBIF 0x14
#define SEPS525_RGB_POL 0x15
#define SEPS525_MEMORY_WRITEMODE 0x16
#define SEPS525_MX1_ADDR 0x17
#define SEPS525_MX2_ADDR 0x18
#define SEPS525_MY1_ADDR 0x19
#define SEPS525_MY2_ADDR 0x1A
#define SEPS525_MEMORY_ACCESS_POINTER_X 0x20
#define SEPS525_MEMORY_ACCESS_POINTER_Y 0x21
#define SEPS525_DDRAM_DATA_ACCESS_PORT 0x22
#define SEPS525_GRAY_SCALE_TABLE_INDEX 0x50
#define SEPS525_GRAY_SCALE_TABLE_DATA 0x51
#define SEPS525_DUTY 0x28
#define SEPS525_DSL 0x29
#define SEPS525_D1_DDRAM_FAC 0x2E
#define SEPS525_D1_DDRAM_FAR 0x2F
#define SEPS525_D2_DDRAM_SAC 0x31
#define SEPS525_D2_DDRAM_SAR 0x32
#define SEPS525_SCR1_FX1 0x33
#define SEPS525_SCR1_FX2 0x34
#define SEPS525_SCR1_FY1 0x35
#define SEPS525_SCR1_FY2 0x36
#define SEPS525_SCR2_SX1 0x37
#define SEPS525_SCR2_SX2 0x38
#define SEPS525_SCR2_SY1 0x39
#define SEPS525_SCR2_SY2 0x3A
#define SEPS525_SCREEN_SAVER_CONTEROL 0x3B
#define SEPS525_SS_SLEEP_TIMER 0x3C
#define SEPS525_SCREEN_SAVER_MODE 0x3D
#define SEPS525_SS_SCR1_FU 0x3E
#define SEPS525_SS_SCR1_MXY 0x3F
#define SEPS525_SS_SCR2_FU 0x40
#define SEPS525_SS_SCR2_MXY 0x41
#define SEPS525_MOVING_DIRECTION 0x42
#define SEPS525_SS_SCR2_SX1 0x47
#define SEPS525_SS_SCR2_SX2 0x48
#define SEPS525_SS_SCR2_SY1 0x49
#define SEPS525_SS_SCR2_SY2 0x4A

/* SEPS525_DISPLAYMODE_SET */
#define MODE_SWAP_BGR	BIT(7)
#define MODE_SM		BIT(6)
#define MODE_RD		BIT(5)
#define MODE_CD		BIT(4)

#define usleep_range(a, b) udelay((b))

/* This should be fixed */
struct seps525_priv {
	struct gpio_desc reset_gpio;
	struct gpio_desc dc_gpio;
	struct spi_slave *slave;
	struct video_priv *priv;
};

struct seps525_priv seps252;

static int seps525_spi_write_cmd(struct spi_slave *spi, u8 reg)
{
	unsigned long flags = SPI_XFER_BEGIN;
	u8 buf8 = reg;
	int ret = 0;

	flags |= SPI_XFER_END;

	dm_gpio_set_value(&seps252.dc_gpio, 0);
	ret = spi_xfer(spi, 8, &buf8, NULL, flags);
	if (ret)
		debug("%s: Failed to send: %d\n", __func__, ret);

	return ret;
}

static int seps525_spi_write_data(struct spi_slave *spi, u8 val)
{
	unsigned long flags = SPI_XFER_BEGIN;
	u8 buf8 = val;
	int ret = 0;

	flags |= SPI_XFER_END;

	dm_gpio_set_value(&seps252.dc_gpio, 1);
	ret = spi_xfer(spi, 8, &buf8, NULL, flags);
	if (ret)
		debug("%s: Failed to send: %d\n", __func__, ret);

	return ret;
}


static void seps525_spi_write(struct spi_slave *spi, u8 reg, u8 val)
{
	seps525_spi_write_cmd(spi, reg);
	seps525_spi_write_data(spi, val);
}

static int seps525_display_init(struct spi_slave *spi)
{
	usleep_range(1000, 5000);

	/* Disable Oscillator Power Down */
	seps525_spi_write(spi, SEPS525_REDUCE_CURRENT, 0x03);
	usleep_range(1000, 5000);
	/* Set Normal Driving Current */
	seps525_spi_write(spi, SEPS525_REDUCE_CURRENT, 0x00);
	usleep_range(1000, 5000);

	seps525_spi_write(spi, SEPS525_SCREEN_SAVER_CONTEROL, 0x00);
	/* Set EXPORT1 Pin at Internal Clock */
	seps525_spi_write(spi, SEPS525_OSC_CTL, 0x01);
	/* Set Clock as 120 Frames/Sec */
	seps525_spi_write(spi, SEPS525_CLOCK_DIV, 0x90);
	/* Set Reference Voltage Controlled by External Resister */
	seps525_spi_write(spi, SEPS525_IREF, 0x01);

	/* precharge time R G B */
	seps525_spi_write(spi, SEPS525_PRECHARGE_TIME_R, 0x04);
	seps525_spi_write(spi, SEPS525_PRECHARGE_TIME_G, 0x05);
	seps525_spi_write(spi, SEPS525_PRECHARGE_TIME_B, 0x05);

	/* precharge current R G B (uA) */
	seps525_spi_write(spi, SEPS525_PRECHARGE_CURRENT_R, 0x9D);
	seps525_spi_write(spi, SEPS525_PRECHARGE_CURRENT_G, 0x8C);
	seps525_spi_write(spi, SEPS525_PRECHARGE_CURRENT_B, 0x57);

	/* driving current R G B (uA) */
	seps525_spi_write(spi, SEPS525_DRIVING_CURRENT_R, 0x56);
	seps525_spi_write(spi, SEPS525_DRIVING_CURRENT_G, 0x4D);
	seps525_spi_write(spi, SEPS525_DRIVING_CURRENT_B, 0x46);
	/* Set Color Sequence */
	seps525_spi_write(spi, SEPS525_DISPLAYMODE_SET, 0x00);
	seps525_spi_write(spi, SEPS525_RGBIF, 0x01); /* Set MCU Interface Mode */
	/* Set Memory Write Mode */
	seps525_spi_write(spi, SEPS525_MEMORY_WRITEMODE, 0x66);
	seps525_spi_write(spi, SEPS525_DUTY, 0x7F); /* 1/128 Duty (0x0F~0x7F) */
	/* Set Mapping RAM Display Start Line (0x00~0x7F) */
	seps525_spi_write(spi, SEPS525_DSL, 0x00);
	seps525_spi_write(spi, SEPS525_DISP_ONOFF, 0x01); /* Display On (0x00/0x01) */
	/* Set All Internal Register Value as Normal Mode */
	seps525_spi_write(spi, SEPS525_SOFT_RST, 0x00);
	/* Set RGB Interface Polarity as Active Low */
	seps525_spi_write(spi, SEPS525_RGB_POL, 0x00);

	seps525_spi_write_cmd(spi, SEPS525_DDRAM_DATA_ACCESS_PORT);

	return 0;
}

static int seps525_spi_startup(struct udevice *dev)
{
	int ret;

	struct spi_slave *slave = dev_get_parent_priv(dev);

	ret = spi_claim_bus(slave);
	if (ret) {
		debug("%s: Failed to claim SPI bus: %d\n", __func__, ret);
		return ret;
	}

	dm_gpio_set_value(&seps252.reset_gpio, 1);
	udelay(1000);
	dm_gpio_set_value(&seps252.reset_gpio, 0);

	seps525_display_init(slave);

	spi_release_bus(slave);

	return 0;
}

static void seps525_sync(struct video_priv *priv, struct udevice *vid, bool force)
{
	int i, ret;
	u8 *start;
	u8 data1, data2;
	printf("Sync with LCD\n");
	ret = spi_claim_bus(seps252.slave);
	if (ret) {
		debug("%s: Failed to claim SPI bus: %d\n", __func__, ret);
		return ret;
	}

	/* start position X,Y */
	seps525_spi_write(seps252.slave, SEPS525_MEMORY_ACCESS_POINTER_X, 0);
	seps525_spi_write(seps252.slave, SEPS525_MEMORY_ACCESS_POINTER_Y, 0);

	seps525_spi_write_cmd(seps252.slave, SEPS525_DDRAM_DATA_ACCESS_PORT);
	start = priv->fb;
	for (i = 0; i < 0xa000/2; i++) {
		data2 = *start++;
		data1 = *start++;
		seps525_spi_write_data(seps252.slave, data1);
		seps525_spi_write_data(seps252.slave, data2);
	}

	spi_release_bus(seps252.slave);
}

static int seps525_probe(struct udevice *dev)
{
	struct video_priv *uc_priv = dev_get_uclass_priv(dev);
	struct spi_slave *slave = dev_get_parent_priv(dev);
	int ret;

	ret = gpio_request_by_name(dev, "reset-gpios", 0,
					&seps252.reset_gpio, GPIOD_IS_OUT);
	if (ret) {
		dev_err(dev, "missing reset GPIO\n");
		return ret;
	}

	ret = gpio_request_by_name(dev, "dc-gpios", 0,
					&seps252.dc_gpio, GPIOD_IS_OUT);
	if (ret) {
		dev_err(dev, "missing dc GPIO\n");
		return ret;
	}

	ret = dm_spi_claim_bus(dev);
	if (ret) {
		dev_err(dev, "Can't claim the bus\n");
		return ret;
	}

	uc_priv->bpix = VIDEO_BPP16;
	uc_priv->xsize = WIDTH;
	uc_priv->ysize = HEIGHT;
	uc_priv->rot = 0;

	seps252.slave = slave;
	seps252.priv = uc_priv;

	seps525_spi_startup(dev);

	video_sync_hook_setup(seps525_sync);

	return 0;
}

static int nx_display_bind(struct udevice *dev)
{
	struct video_uc_platdata *plat = dev_get_uclass_platdata(dev);

	plat->size = WIDTH * HEIGHT * 16;

	return 0;
}

static const struct udevice_id seps525_ids[] = {
	{ .compatible = "syncoam,seps525" },
	{ }
};

U_BOOT_DRIVER(mvebu_video) = {
	.name	= "seps525_video",
	.id	= UCLASS_VIDEO,
	.of_match = seps525_ids,
	.platdata_auto_alloc_size = sizeof(struct video_uc_platdata),
	.bind = nx_display_bind,
	.probe	= seps525_probe,
	.priv_auto_alloc_size = sizeof(struct video_priv),
};

static int do_seps525_access(struct cmd_tbl_s *cmdtp, int flag, int argc,
			     char * const argv[])
{
	int ret;
	u8 cmd, data;

	cmd = simple_strtoul(argv[2], NULL, 8);
	data = simple_strtoul(argv[3], NULL, 8);

	ret = spi_claim_bus(seps252.slave);
	if (ret) {
		debug("%s: Failed to claim SPI bus: %d\n", __func__, ret);
		return ret;
	}

	seps525_spi_write(seps252.slave, cmd, data);

	spi_release_bus(seps252.slave);

	return 0;
}

static int do_seps525_mem(struct cmd_tbl_s *cmdtp, int flag, int argc,
			     char * const argv[])
{
	u16 xs, ys; 
	unsigned short data;
	u8 data1, data2;
	int i, ret;
	u8 *start;

	xs = simple_strtoul(argv[2], NULL, 16);
	ys = simple_strtoul(argv[3], NULL, 16);

	ret = spi_claim_bus(seps252.slave);
	if (ret) {
		debug("%s: Failed to claim SPI bus: %d\n", __func__, ret);
		return ret;
	}

	/* start position X,Y */
	seps525_spi_write(seps252.slave, SEPS525_MEMORY_ACCESS_POINTER_X, 0);
	seps525_spi_write(seps252.slave, SEPS525_MEMORY_ACCESS_POINTER_Y, 0);

	seps525_spi_write_cmd(seps252.slave, SEPS525_DDRAM_DATA_ACCESS_PORT);
	start = seps252.priv->fb;
	for (i = 0; i < xs; i++) {
		data2 = *start++;
		data1 = *start++;
		seps525_spi_write_data(seps252.slave, data1);
		seps525_spi_write_data(seps252.slave, data2);
	}

	spi_release_bus(seps252.slave);

	return 0;
}


void main_seps();

static int do_seps525_main(struct cmd_tbl_s *cmdtp, int flag, int argc,
			     char * const argv[])
{
	int ret;

	ret = spi_claim_bus(seps252.slave);
	if (ret) {
		printf("%s: Failed to claim SPI bus: %d\n", __func__, ret);
		return ret;
	}

	main_seps();

	spi_release_bus(seps252.slave);

	return 0;
}

static struct cmd_tbl_s cmd_zynqmp_sub[] = {
	U_BOOT_CMD_MKENT(display, 4, 0, do_seps525_access, "", ""),
	U_BOOT_CMD_MKENT(main, 4, 0, do_seps525_main, "", ""),
	U_BOOT_CMD_MKENT(mem, 4, 0, do_seps525_mem, "", ""),
};

static int do_seps525(struct cmd_tbl_s *cmdtp, int flag, int argc,
		     char *const argv[])
{
	struct cmd_tbl_s *c;

	if (argc < 2)
		return CMD_RET_USAGE;

	c = find_cmd_tbl(argv[1], &cmd_zynqmp_sub[0],
			 ARRAY_SIZE(cmd_zynqmp_sub));

	if (c)
		return c->cmd(c, flag, argc, argv);
	else
		return CMD_RET_USAGE;
}

U_BOOT_CMD(
	seps525,	9,	1,	do_seps525,
	"set seps525display",
	""
);

void Delay(u32 sec)
{
	mdelay(1000 * sec);
};

#define uDelay udelay

void Write_Command(unsigned char Data) {
	seps525_spi_write_cmd(seps252.slave, Data);
}


void Write_Data(unsigned char Data) {
	seps525_spi_write_data(seps252.slave, Data);
}

#include "seps525-standalone.c"
